/* Conceptos:
 - Critical section
 - Synchronization mechanism
 - Mutex
 - Blocking operations
 */
#define _REENTRANT 0

#include <pthread.h>
/* all mutex operations return 0 on success */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int shared=0;

/*
 * Static initialisation of mutex
 * No attributes can be given (see later).
 * Similar to calling
 * int pthread_mutex_init(pthread_mutex_t *, const pthread_mutexattr_t *);
 * as follows: 
 * pthread_mutex_init(& shared_mutex, NULL);
 */
pthread_mutex_t shared_mutex = PTHREAD_MUTEX_INITIALIZER;

void *my_run(void *arg){
	int i;
	extern int shared;
	
	for(i=*(int *)arg; i>0; i--){
		/*
		 * Blocks until lock is acquired
		 */
		pthread_mutex_lock(&shared_mutex);
		int temp = shared;
		putchar('+');
		shared = temp + 1;
		pthread_mutex_unlock(&shared_mutex);
	}
	
	return NULL;
}

int main(){
	pthread_t created;
	extern int shared;
	extern pthread_mutex_t shared_mutex;
	int counter = 10000;

	int flag = pthread_create(&created, NULL, &my_run, (void *)&counter);
	
	if(flag != 0){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}
	int i;
	for(i = counter; i>0; i--){
		putchar('-');
		pthread_mutex_lock(&shared_mutex);
		int temp = shared;
		putchar('-');
		shared = temp - 1;
		pthread_mutex_unlock(&shared_mutex);
	}
	
	void *resval;
	flag = pthread_join(created,&resval);
	if(flag){
		fprintf(stderr,"Thread join error\n");
		return(EXIT_FAILURE);
	}
	pthread_mutex_destroy(&shared_mutex);
	
	printf("Shared = %d",shared);	
	return (EXIT_SUCCESS);
}