/* Conceptos:
 - Bounded buffer
 - Semaphore
 - Atomic operations
 */
#define _REENTRANT 0

#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * semaphore operations
 * not supported on mac os x!
 * these return 0 on success
 *
 * Not seen in this code is
 * int sem_trywait(sem_t *sem);
 * int sem_getvalue(sem_t *sem, int *sval);
 *
 */

#define BUFSIZE 5

struct Buffer {
	char buf[BUFSIZE];
	sem_t freespace, usedspace;
	int in, out;
};

struct Buffer *makeBuffer(){
	struct Buffer *result;	
	int resultcode;
	
	result = (struct Buffer *)malloc(sizeof(struct Buffer));

	if(!result){
		fprintf(stderr,"Buffer creation error\n");
		exit(EXIT_FAILURE);
	}

	/*
	 * Initialise counting semaphore.
	 * Second argument always 0 on linux.
	 * Third argument = the count of the semaphore
	 */
	/*
	 *  &(result->freespace) == &result->freespace
	 */	
	resultcode = sem_init(&result->freespace,0,BUFSIZE);
	if(resultcode){
		fprintf(stderr,"Buffer creation error (sem_init freespace)\n");
		exit(EXIT_FAILURE);
	}
	
	resultcode = sem_init(&result->usedspace,0,0);
	if(resultcode){
		fprintf(stderr,"Buffer creation error (sem_init usedspace)\n");
		exit(EXIT_FAILURE);
	}
	
	result->in = 0;
	result->out = 0;
	
	return result;
}

void freeBuffer(struct Buffer *buf){
	/*
	 * Destroy semaphores after use
	 */
	sem_destroy(&buf->freespace);
	sem_destroy(&buf->usedspace);
	free(buf);
}

void putcharacter(struct Buffer *buf, char c){
	/*
	 * waits if = 0, - atomically
	 */
	sem_wait(&buf->freespace);
	buf->buf[buf->in] = c;
	buf->in=(buf->in+1)%BUFSIZE;
	/*
	 * + atomically
	 */
	sem_post(&buf->usedspace);
}

char getcharacter(struct Buffer *buf){
	char result;
	sem_wait(&buf->usedspace);
	result = buf->buf[buf->out];
	buf->out = (buf->out+1)%BUFSIZE;
	sem_post(&buf->freespace);
	return result;
}

void *printer(void *arg){
	struct Buffer *buf = (struct Buffer *) arg;
	
	while(1){
		int i;
		for(i=0; i<7; i++)
			putchar(getcharacter(buf));
		putchar('\n');
		sleep(1);
	}
	return NULL;
}

int main(){
	pthread_t created;
	struct Buffer *buf = makeBuffer();
	
	int flag = pthread_create(&created, NULL, &printer, (void *)buf);
	if(flag){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}
	
	char *toprint="The Hitchiker's Guide to the Galaxy.\n";
	int i;
	for(i = 0; i<strlen(toprint); i++){
		putcharacter(buf,toprint[i]);
		printf("+");
	}
	sleep(5);
	
	freeBuffer(buf);
	
	return(EXIT_SUCCESS);
}
