/* Conceptos:
 - Thread Attributes
 - Detached threads
 */
#define _REENTRANT 0

#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * semaphore operations
 * not supported on mac os x!
 * these return 0 on success
 */

#define BUFSIZE 5

struct Buffer {
        char buf[BUFSIZE];
        sem_t freespace, usedspace;
        int in, out;
};

int consumer_done = 0;

struct Buffer *makeBuffer(){
        struct Buffer *result;
        int resultcode;

        result = (struct Buffer *)malloc(sizeof(struct Buffer));

        if(!result){
                fprintf(stderr,"Buffer creation error\n");
                exit(EXIT_FAILURE);
        }

        resultcode = sem_init(&result->freespace,0,BUFSIZE);
        if(resultcode){
                fprintf(stderr,"Buffer creation error\n");
                exit(EXIT_FAILURE);
        }

        resultcode = sem_init(&result->usedspace,0,0);
        if(resultcode){
                fprintf(stderr,"Buffer creation error\n");
                exit(EXIT_FAILURE);
        }

        result->in = 0;
        result->out = 0;

        return result;
}

void freeBuffer(struct Buffer *buf){
        sem_destroy(&buf->freespace);
        sem_destroy(&buf->usedspace);
        free(buf);
}

void putcharacter(struct Buffer *buf, char c){
        sem_wait(&buf->freespace);
        buf->buf[buf->in] = c;
        buf->in=(buf->in+1)%BUFSIZE;
        sem_post(&buf->usedspace);
}

char getcharacter(struct Buffer *buf){
        char result;
        sem_wait(&buf->usedspace);
        result = buf->buf[buf->out];
        buf->out = (buf->out+1)%BUFSIZE;
        sem_post(&buf->freespace);
        return result;
}

void *printer(void *arg){
        struct Buffer *buf = (struct Buffer *) arg;

        char toprint;
        extern int consumer_done;
        
        while(1){
                int i;
                for(i=0; i<7; i++){
                		toprint = getcharacter(buf);
                        if(toprint == EOF){
                                printf("\nEnd of Input");
                                consumer_done = 1;
                                return NULL;
                        }
                        else
                        	putchar(toprint);
                }
                putchar('\n');
                sleep(1);
        }
        return NULL;
}

int main(){
        pthread_t created;
        struct Buffer *buf = makeBuffer();
        /*
         * Interesting attributes:
         * joinable or not
         * scheduling policy attributes
         */
        pthread_attr_t attributes;
        int flag = pthread_attr_init(&attributes);
        if(flag){
                fprintf(stderr,"Thread attributes creation error\n");
                return(EXIT_FAILURE);
        }
        /*
         * Default is PTHREAD_CREATE_JOINABLE
         */
        flag = pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_DETACHED);
        if(flag){
                fprintf(stderr,"Thread attributes set error\n");
                return(EXIT_FAILURE);
        }

        flag = pthread_create(&created, &attributes, &printer, (void *)buf);
        if(flag){
                fprintf(stderr,"Thread creation error\n");
                return(EXIT_FAILURE);
        }
        /*
         * Dont need these anymore
         */
        pthread_attr_destroy(&attributes);

        char *toprint="The Hitchiker's Guide to the Galaxy.\n";
        int i;
        for(i = 0; i<strlen(toprint); i++){
                putcharacter(buf,toprint[i]);
                printf("+");
        }
        putcharacter(buf,EOF);
 
        extern int consumer_done;
        while(!consumer_done){
        		printf("*");
                sleep(1);
        }

        freeBuffer(buf);

        return(EXIT_SUCCESS);
}
