/* Conceptos:
 - Thread Canceling
 */
#define _REENTRANT 0
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

void *threadcode(void *arg){
	/*
	 * First arg PTHREAD_CANCEL_ENABLE  or PTHREAD_CANCEL_DISABLE
	 * Second = pointer to where old state should be written (NULL possible)
	 */
	int oldval;
	int flag = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldval);
	if(flag) {
		fprintf(stderr,"Thread cancel state set error\n");
		exit(EXIT_FAILURE);
	}
	if (oldval == PTHREAD_CANCEL_ENABLE)
		printf("Standard cancelation enabled\n");
	else
		printf("Standard cancelation disabled\n");

	/*
	 * First arg PTHREAD_CANCEL_DEFERRED or PTHREAD_CANCEL_ASYNCHRONOUS
	 * Second = pointer to where old state should be written (NULL possible)
	 *
	 * If deferred, cancels occur on blocking operations (+-)  + pthread_testcancel();
	 * (First depends on adherence to specifications)
	 */
	flag = pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oldval);
	if(flag) {
		fprintf(stderr,"Thread cancel type set error\n");
		exit(EXIT_FAILURE);
	}
	if (oldval == PTHREAD_CANCEL_DEFERRED)
		printf("Standard cancelation deferred\n");
	else
		printf("Standard cancelation immediate\n");
	
	
	int i;
	for(i=10; i>0; i--){
		printf("Thread alive.\n");
		sleep(1);
		/* pthread_testcancel(); */
	}

	printf("Thread survived!\n");
	return NULL;
}

int main(){
	pthread_t created;
	int counter = 10000;
	int flag = pthread_create(&created, NULL, &threadcode, (void *)&counter);
	
	if(flag != 0){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}

	sleep(3);
	
	printf("Going for the kill!\n");
	flag = pthread_cancel(created);
	if(flag){
		fprintf(stderr,"Thread cancel error\n");
		return(EXIT_FAILURE);
	}
	printf("Cancel sent\n");
	printf("Waiting for join\n");
	flag = pthread_join(created,NULL);
	if(flag){
		fprintf(stderr,"Thread join error\n");
		return(EXIT_FAILURE);
	}
	
	return (EXIT_SUCCESS);
}