
#define _REENTRANT 0
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXSTACK 100

typedef struct {
	int top;
	void **contents;
	pthread_mutex_t read_write;
	pthread_mutex_t empty;
} Stack;

Stack *makeStack(){
	Stack *st = (Stack *)malloc(sizeof(Stack));
	if(!st)
		exit(EXIT_FAILURE);
	
	void **contents = (void **)malloc(sizeof(void*)*MAXSTACK);
	if(!contents)
		exit(EXIT_FAILURE);
	
	st->top = 0;
	st->contents = contents;
	int fail;
	/* returns 0 on success! */
	/*
	 * Second parameter: mutex attributes
	 * not supported by all POSIX implementations
	 * we leave it NULL
	 */
	fail = pthread_mutex_init(&st->read_write, NULL);
	fail = fail || pthread_mutex_init(&st->empty, NULL); 
	
	if(fail)
		exit(EXIT_FAILURE);
		
	pthread_mutex_lock(&st->empty);
	return st;
}

void freeStack(Stack *st){
	free(st->contents);
	pthread_mutex_destroy(&st->read_write);
	pthread_mutex_destroy(&st->empty);
	free(st);
}

void *top(Stack *st){
	pthread_mutex_lock(&st->empty);
	pthread_mutex_lock(&st->read_write);
	void *result = *(st->contents + st->top);
	pthread_mutex_unlock(&st->read_write);
	pthread_mutex_unlock(&st->empty);
	return result;
}

void push(Stack *st, void *arg){
	pthread_mutex_lock(&st->read_write);
	*(st->contents + ++st->top) = arg;
	if(st->top == 1)
		pthread_mutex_unlock(&st->empty);
	pthread_mutex_unlock(&st->read_write);
}

void *pop(Stack *st){
	pthread_mutex_lock(&st->empty);
	pthread_mutex_lock(&st->read_write);
	void *result = *(st->contents + st->top--);
	if(st->top != 0)
		pthread_mutex_unlock(&st->empty);
	pthread_mutex_unlock(&st->read_write);
	return result;
}

void *popper(void *arg){
	Stack *st = (Stack *)arg;
	while(1){
		putchar(*(char *)pop(st));
		putchar('\n');
	}
	return NULL;
}

int main(){
	pthread_t created;
	Stack *st = makeStack();

	int flag = pthread_create(&created, NULL, &popper, (void *)st);
	if(flag){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}
	char *toprint="The Hitchiker's Guide to the Galaxy.\n";
	int i;
	for(i = 0; i<strlen(toprint)/2; i++){
		push(st,(void *)toprint+i);
		printf("+\n");
	}
	printf("|\n");
	sleep(5);
	printf("|\n");
	for(; i<strlen(toprint); i++){
		push(st,(void *)toprint+i);
		printf("+\n");
		sleep(1);
	}
	
	/*
	 * How do you make this clean?
	 */
	sleep(2);
	
	freeStack(st);
	
	return(EXIT_SUCCESS);
}
