
#include "07_monitor.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define MAXSTACK 100

typedef struct {
	int top;
	void **contents;
	Monitor *mon;
} Stack;

Stack *makeStack(){
	Stack *st = (Stack *)malloc(sizeof(Stack));
	if(!st)
		exit(EXIT_FAILURE);
	
	void **contents = (void **)malloc(sizeof(void*)*MAXSTACK);
	if(!contents)
		exit(EXIT_FAILURE);
	
	st->top = 0;
	st->contents = contents;
	st->mon = createMonitor();
	if(!st->mon)
		exit(EXIT_FAILURE);
	return st;
}

void freeStack(Stack *st){
	free(st->contents);
	destroyMonitor(st->mon);
	free(st);
}

void push(Stack *st, void *arg){
	monitorEnter(st->mon);
	*(st->contents + ++st->top) = arg;
	monitorNotifyWaiting(st->mon);
	monitorExit(st->mon);
}

void *pop(Stack *st){
	monitorEnter(st->mon);
	/*
	 * Always check monitor condition variable with while loop!
	 */
	while(st->top == 0)
		monitorWaitOn(st->mon);
	void *result = *(st->contents + st->top--);
	monitorExit(st->mon);
	return result;
}

void *popper(void *arg){
	Stack *st = (Stack *)arg;
	while(1){
		putchar(*(char *)pop(st));
		putchar('\n');
	}
	return NULL;
}

int main(){
	pthread_t created;
	Stack *st = makeStack();
	
	int flag = pthread_create(&created, NULL, &popper, (void *)st);
	if(flag){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}
	char *toprint="The Hitchiker's Guide to the Galaxy.\n";
	int i;
	for(i = 0; i<strlen(toprint)/2; i++){
		push(st,(void *)toprint+i);
		printf("+\n");
	}
	printf("|\n");
	sleep(5);
	printf("|\n");
	for(; i<strlen(toprint); i++){
		push(st,(void *)toprint+i);
		printf("+\n");
		sleep(1);
	}

	sleep(2);
	
	freeStack(st);
	
	return(EXIT_SUCCESS);
}
