/* Conceptos:
 - Condition Variables
 */
#define _REENTRANT 0
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define MAXSTACK 6

typedef struct {
	int top;
	void **contents;
	pthread_mutex_t read_write;
	pthread_cond_t condvar;
} Stack;

Stack *makeStack(){
	Stack *st = (Stack *)malloc(sizeof(Stack));
	if(!st)
		exit(EXIT_FAILURE);
	
	void **contents = (void **)malloc(sizeof(void*)*MAXSTACK);
	if(!contents)
		exit(EXIT_FAILURE);
	
	st->top = 0;
	st->contents = contents;
	int fail;
	fail = pthread_mutex_init(&st->read_write, NULL);
	/* returns 0 on success! */
	/*
	 * Second argument are condition variable attributes
	 * Only used with processes. We use NULL here
	 *
	 * Static initialisation:
	 * pthread_cond_t myconvar = PTHREAD_COND_INITIALIZER;
	 */
	fail = fail || pthread_cond_init(&st->condvar,NULL);
	
	if(fail)
		exit(EXIT_FAILURE);
	
	return st;
}

void freeStack(Stack *st){
	free(st->contents);
	pthread_mutex_destroy(&st->read_write);
	pthread_cond_destroy(&st->condvar);
	free(st);
}

void push(Stack *st, void *arg){
	pthread_mutex_lock(&st->read_write);
	while(st->top == MAXSTACK)
		pthread_cond_wait(&st->condvar,&st->read_write);
	*(st->contents + ++st->top) = arg;
	/*
	 * Also exists a pthread_cond_signal
	 */
	pthread_cond_broadcast(&st->condvar);
	pthread_mutex_unlock(&st->read_write);
}

void *pop(Stack *st){
	pthread_mutex_lock(&st->read_write);
	while(st->top == 0)
		pthread_cond_wait(&st->condvar,&st->read_write);
	void *result = *(st->contents + st->top--);
	pthread_cond_broadcast(&st->condvar);
	pthread_mutex_unlock(&st->read_write);
	return result;
}

void *popper(void *arg){
	Stack *st = (Stack *)arg;
	while(1){
		putchar(*(char *)pop(st));
		putchar('\n');
	}
	return NULL;
}

int main(){
	pthread_t created;
	Stack *st = makeStack();
	
	int flag = pthread_create(&created, NULL, &popper, (void *)st);
	if(flag){
		fprintf(stderr,"Thread creation error\n");
		return(EXIT_FAILURE);
	}
	char *toprint="The Hitchiker's Guide to the Galaxy.\n";
	int i;
	for(i = 0; i<strlen(toprint)/2; i++){
		push(st,(void *)toprint+i);
		printf("+\n");
	}
	printf("|\n");
	sleep(5);
	printf("|\n");
	for(; i<strlen(toprint); i++){
		push(st,(void *)toprint+i);
		printf("+\n");
		sleep(1);
	}
	
	sleep(2);
	
	freeStack(st);
	
	return(EXIT_SUCCESS);
}
